// @ myth - server - monster.h
//
// Myth is a multiuser screen oriented adventure game.
// Copyright (C) 1989 Per Abrahamsen
// 
// This file is part of Myth.
// 
// Myth is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 1, or (at your option) any later
// version.
// 
// Myth is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Myth; see the file COPYING.  If not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// My email address is: amanda@iesd.auc.dk

// @@ begin

#ifndef MONSTER_HEADER
#define MONSTER_HEADER

// @@ include

#include "server/object.h"
#include "server/pack.h"
#include "server/body.h"
#include "common/command.h"
#include "common/string.h"
#include "server/global.h"

struct PARSE;
struct LEVEL;
struct SCORE_ENTRY;

// @@ MONSTER

class MONSTER : public OBJECT
{
    // Inventory and Body
    //
protected:
    BODY* body;
public:
    PACK pack;			// Contain all the monsters inventory
    OBJECT* weapon();		// Current wielded weapon
    long size()
    { return body->size() + pack.size(); }
    long weight()
    { return body->weight() + pack.weight(); }
    // Feed the monster (Let it eat)
    cont feed(int value, force flag)
    { return body->feed(value, flag); }
    void breathe(LEVEL* level, coor pos)
    { body->breathe(level, this, pos); }
    //
    // Move and Fight Points
    //
private:
    int MovePoint;		// Number of moves left
    int FightPoint;		// Number of violent moves left
    int speed()			// How fast is it?
    { return body->speed(); }
public:
    bool MayFight()		// Is monster allowed to fight?
    { return FightPoint > 0; }
    bool MayMove()		// Is monster allowed to move?
    { return !dead() && MovePoint > 0; }
    void UseFightPoint(int point) // Spend some fight points
    { FightPoint -= point; }
    void UseMovePoint(int point) // Spend some move points
    { MovePoint -= point; }
    virtual void TickFight()	// Allow monster to fight
    { FightPoint = min(speed(), FightPoint + speed()); }
    virtual void TickMove()	// Allow monster to move
    { MovePoint = min(speed(), MovePoint + speed()); }
    //
    // Moving around
    //
private:
    coor where;			// Current position of monster
protected:
    level_type DungeonLevel;	// Level where monster is (going to be)
    int NewLocation;		// Position on new level
public:
    coor position()		// Where is the monster on the level
    { return where; }
    bool IsConfused();		// Is the monster confused? (chance)
    level_type depth()		// On which level is the monster
    { return DungeonLevel; }
    bool done()			// Is monster deleted from the game?
    { return DungeonLevel == LEVEL_REMOVE; }
    bool escape()		// Has it climbed out off the dungeon?
    { return DungeonLevel == LEVEL_EXIT; }
    void NewLevel(level_type depth, char where);
    virtual void enter(LEVEL* level); // New level
    void leave(LEVEL* level);	//  Leave a bunch if you appear in it.
    void teleport(LEVEL* level, int length, holy_type holy = NORMAL);
    virtual void move(LEVEL* level, coor pos); // New position
    //
    // Death
    //
protected:
    STRING rip;			// Why is it death?
public:
    virtual void kill(STRING why, LEVEL* level);
    void kill(const char* why, LEVEL* level) // Kill monster, and say why
    { kill(STRING(why), level); }
    bool dead()			// Still alive?
    { return body->dead(); }
    cont event(LEVEL*, coor, MONSTER* = NULL)
    { return body->done() ? DONE : CONTINUE; }
    SCORE_ENTRY* ScoreEntry();
    //
    // Identification
    //
protected:
    char MyLetter;		// What inventory letter is used
public:
    char letter()		// Letter in inventory
    { return MyLetter; }
    void SetLetter(char letter)	// Select letter
    { MyLetter = letter; }
    virtual object_type LookType()
    { return body->LookType(); }
    symbol_type symbol()
    { return dead() ? SYMBOL_FOOD : SYMBOL_ERROR; }
    bool IsFood()
    { return dead(); }
    char look();
    STRING print(print_flag flag, long count, PLAYER* player);
    virtual const char* LivingName()
    { return body->BodyName(); }
    const char* name()
    { return body->name(); }
    const char* PluralName()
    { return body->PluralName(); }
    const char* ClassName()
    { return body->ClassName(); }
    const char* PluralClassName()
    { return body->PluralClassName(); }
    bool IsMonster(force flag = CHECK) // Identify this as a monster
    { return flag == FORCE || !dead(); }
    bool visible();
    comp compare(OBJECT* object); // Monsters is never equal
    bool MightHit()		// Monsters may normally be hit
    { return dead() ? OBJECT::MightHit() : TRUE; }
    long worth(MONSTER*)
    { return dead() ? 25L : 0L; }
    bool CanRemove(MONSTER*)
    { return dead(); }
    virtual int dexterity()
    { return 10; }
    //
    // Abilities
    //
    bool CheckAbility(ability_type ability) // Add pack here!
    { return body->CheckAbility(ability) ||
	  pack.CheckAbility(ability); }
    int GetValue(ability_type ability) // Add pack here!
    { return body->GetValue(ability) + pack.GetValue(ability); }
    void AddAbility(COUNTER* counter)
    { body->AddAbility(counter); }
    void RemoveAbility(ability_type ability, force flag)
    { body->RemoveAbility(ability, flag); }
    void SetVisible()
    { RemoveAbility(ABILITY_INVISIBLE, FORCE); } 
    void ClearVisible()
    { AddAbility(new COUNTER_INVISIBLE); }
    bool SeeInvisible()		// Can monster see invisible objects?
    { return CheckAbility(ABILITY_SEE_INVISIBLE); }
    bool light()		// Does it carry a light source?
    { return body->CheckAbility(ABILITY_LIGHT) || pack.light(); }
    //
    // Strength, Health && Experience
    //
protected:
    int MyExperience;
    long MyExpPoint;
    long MyScore;
public:
    int strength();		// Monsters strength
    void AddStrength(int str)
    { body->AddStrength(str); }
    void RestoreStrength()
    { body->RestoreStrength(); }
    bool CanMove(long extra = 0L); // Collapsed?
    bool HeavyLoad(long extra = 0L); // Carrying to much?
    bool LightLoad(long extra = 0L); // Carrying very little?
    bool CanFight();		// Acid blob's can't fight
    void heal(int hp)		// Heal monster
    { body->heal(hp); }
    void RestoreHealth()
    { body->RestoreHealth(); }
    int experience()		// Normally 1 to 25
    { return MyExperience; }
    void GainExperience(int exp);
    void SetExperience(int exp);
    void GainScore(int score)
    { MyScore += score; }
    int ToDamage()		// How much extra damage when hit
    { return body->damage(); }
    //
    // Use
    //
private:
    void eat(LEVEL*, MONSTER* monster); // Eat the dead body
protected:
    void SpendTime(LEVEL* level); // Update counters etc.
public:
    virtual bool IsEnemy(OBJECT*) // TRUE, iff monster is an enemy
    { return FALSE; }
    virtual bool IsWanted(OBJECT*) // TRUE, iff monster want this item
    { return FALSE; }
    virtual bool IsPicking(OBJECT*) // TRUE, iff monster pick this up
    { return FALSE; }
    void apply(LEVEL*, MONSTER* monster); // Eat the dead body
    bool MayWearRing()		// Any free ringfinger?
    { return body->MaxRing() > pack.CurrentRing(); }
    QUESTION* DoApply(LEVEL* level, PLAYER* player, CHANNEL* channel);
    virtual cont play(LEVEL* level); // Move monster
    bool AcceptEnter(MONSTER*);	// Don't accept other monsters here
    virtual MONSTER* enemy(LEVEL*) // Find an enemy on this level
    { return this; /* Self destruction! */ }
    cont defend(MONSTER* monster, LEVEL* level); // Start a fight
    bool catch(int power, OBJECT* object);
    bool hit(int power, OBJECT* weapon, MONSTER* attack);
    virtual STRING hurt(int power, OBJECT* weapon,
			MONSTER* attack, LEVEL* level);
    virtual void sleep(LEVEL*, long l); // Tired
    virtual bool sleeping()	// Check if monster is sleeping
    { return FALSE; }
    virtual int knowledge(OBJECT*) // Knowledge of an object type
    { return 0; }
    virtual bool know(OBJECT*) // Is this object identified?
    { return FALSE; }
    virtual void learn(OBJECT*, int value = 1) // Learn about an object type
    { GainExperience(value); }
    virtual void identify(OBJECT*) // Identify an instance af an object
    { GainExperience(5); }
    virtual void forget(int chance) // Forget about a type of objects
    { GainExperience(-100 / chance); }
    OBJECT* polymorph(LEVEL*);	// Change it's body (or turn undead!)
    void InsertAbility(MONSTER* monster) // Carrying a dead monster
    { body->InsertAbility(monster); }
    //
    // Administration
    //
private:
    void create();		// Set default values
public:
    virtual void message(STRING) { }
    virtual void message(const char*) { }
    virtual void SendStatus(status_type, status_value) { }
    void save(BINARY* binary);
    MONSTER(BINARY* binary);
    MONSTER(PARSE* parse, coor pos);
    MONSTER();
    ~MONSTER();
};

// @@ end

#endif MONSTER_HEADER

// @@ emacs
//
// Local Variables:
// mode: c++
// outline-regexp: "// @+ "
// eval: (outline-minor-mode 1)
// End:
