// @ myth - server - object.h
//
// Myth is a multiuser screen oriented adventure game.
// Copyright (C) 1989 Per Abrahamsen
// 
// This file is part of Myth.
// 
// Myth is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 1, or (at your option) any later
// version.
// 
// Myth is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Myth; see the file COPYING.  If not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// My email address is: amanda@iesd.auc.dk

// @@ begin

#ifndef OBJECT_HEADER
#define OBJECT_HEADER

// @@ include

#include "common/common.h"
#include "common/symbol.h"
#include "common/string.h"
#include "server/id.h"

struct BINARY;
struct PARSE;
struct OBJECT;
struct MONSTER;
struct PLAYER;
struct LEVEL;
struct QUESTION;
struct CHANNEL;

// @@ Print

// @@@ enum print_flag

enum print_flag
{
    // Flags
    PRINT_NAME = 0, PRINT_ENTRY = 1, PRINT_COUNT = 2,
    PRINT_VALUE = 4, PRINT_TYPE = 8,  PRINT_WORN = 16,
    PRINT_UPPER = 32, PRINT_THE = 64, PRINT_FORCE = 128,
    // Logic names
    PRINT_OBJECT = PRINT_COUNT | PRINT_VALUE | PRINT_TYPE,
    PRINT_LOOK = PRINT_VALUE | PRINT_TYPE | PRINT_WORN,
    PRINT_LIST = PRINT_ENTRY | PRINT_OBJECT | PRINT_WORN,
    PRINT_MONSTER = PRINT_TYPE | PRINT_THE
};

// @@@ const char* PrintCount(long count)

extern const char* PrintCount(long count);

// @@@ STRING print(OBJECT* object,

extern STRING PrintObject(OBJECT* object, print_flag flag,
			  PLAYER* player = NULL, long count = ERROR);

extern const char* print(OBJECT* object, print_flag flag,
			 PLAYER* player = NULL, long count = ERROR);

// @@ enum holy_type

enum holy_type { CURSED, NORMAL, BLESSED };

// @@ enum damage_type
//
// Bitmap specifying the type of damage.

enum damage_type
{
    DAMAGE_HARMLESS = 0,
    DAMAGE_HIT = 1,
    DAMAGE_CUT = 2,
    DAMAGE_STING = 4,
    DAMAGE_PHYSICAL =  DAMAGE_HIT | DAMAGE_CUT | DAMAGE_STING,
    DAMAGE_FLAME = 8,
    DAMAGE_COLD = 16,
    DAMAGE_MAGIC = 32,
    DAMAGE_LIGHT = 64,
    DAMAGE_ACID = 128,
    DAMAGE_POISON = 256,
    DAMAGE_STONE = 512,
    DAMAGE_DRAIN_LIFE = 1024,
    DAMAGE_END
};

// @@ enum object_type
//
// Warning: objects with the same symbol must be grouped together.
//          Note especially dead monsters and food.

enum object_type
{
    // Special objects
    OBJECT_STAIR,
    // Symbol '!'
    POTION_OF_WATER,
    POTION_OF_STRENGTH,
    POTION_OF_HEALING,
    POTION_OF_EXPERIENCE,
    POTION_OF_SIGHT,
    POTION_OF_ENERGY,
    // Symbol '?'
    SCROLL_OF_BLANK_PAPER,
    SCROLL_OF_KNOWLEDGE,
    SCROLL_OF_PRAYERS,
    SCROLL_OF_MAGIC,
    SCROLL_OF_TELEPORT,
    SCROLL_OF_MAPPING,
    // Symbol '/'
    WAND_TORCH,
    WAND_OF_SLEEP,
    WAND_OF_LIFE,
    WAND_OF_TELEPORT,
    WAND_OF_VISIBILITY,
    WAND_OF_MISSILE,
    WAND_OF_LIGHT,
    // Symbol '='
    RING_OF_ADORNMENT,
    RING_OF_HEALTH,
    RING_OF_TELEPORTATION,
    RING_OF_VISION,
    RING_OF_STRENGTH,
    RING_OF_FIRE_RESISTANCE,
    RING_OF_COLD_RESISTANCE,
    AMULET_OF_YENDOR,
    // Symbol '('
    WEAPON_STONE,
    WEAPON_BOW,
    WEAPON_ARROW,
    WEAPON_DAGGER,
    WEAPON_MACE,
    WEAPON_LONG_SWORD,
    WEAPON_TWO_HANDED_SWORD,
    // Symbol '['
    ARMOR_LEATHER,
    ARMOR_RING,
    ARMOR_PLATE,
    // Symbol '%'
    FOOD_RATION,
    FOOD_FRUIT,
    // Nothing between food and monsters!
    MONSTER_PLAYER,
    MONSTER_WIZARD,
    MONSTER_ACID_BLOB,
    MONSTER_BAT,
    MONSTER_RAT,
    MONSTER_GOBLIN,
    MONSTER_HOBGOBLIN,
    MONSTER_ORC,
    MONSTER_DRAGON,
    MONSTER_LIZARD,
    MONSTER_COCKATRICE,
    MONSTER_WRAITH,
    MONSTER_ELF,
    MONSTER_TROLL,
    MONSTER_FUNGUS,
    // The End
    OBJECT_END,
    OBJECT_ERROR = ERROR
};

extern object_type ObjectToType(const char* string);
extern OBJECT* CreateThing(int value);

// @@ OBJECT

class OBJECT
{
    // Identification
    //
    id_type id;			// Unique identity of object
public:
    id_type identity()		// The id of the object
    { return id; }
    virtual object_type type()	// What kind of object is this?
    { return OBJECT_ERROR; }
    virtual object_type LookType() // What kind of object does it look like?
    { return type(); }
    virtual symbol_type symbol() // What symbol does this use?
    { return SYMBOL_ERROR; }
    virtual char letter()	// What letter is it assigned.
    { return SYMBOL_BUNCH; }
    virtual char look()		// What does it look like?
    { return symbol(); }
    virtual int SpecialBonus()	// Special bonus for a property
    { return 0; }
    virtual bool IsBodyPart()	// Is it a part of a monster body?
    { return FALSE; }
    virtual bool IsMonster(force = CHECK) // Is it a monster? 
    { return FALSE; }
    virtual bool IsPlayer()	// And a player?
    { return FALSE; }
    virtual bool IsThing()	// Is it a thing?
    { return FALSE; }
    virtual bool IsMagic()	// Or a wand, potion, scroll or ring?
    { return FALSE; }
    virtual bool IsWand()
    { return FALSE; }
    virtual bool IsPotion()
    { return FALSE; }
    virtual bool IsScroll()
    { return FALSE; }
    virtual bool IsRing()
    { return FALSE; }
    virtual bool IsWeapon()
    { return FALSE; }
    virtual bool IsArmor()
    { return FALSE; }
    virtual bool IsFood()
    { return FALSE; }
    virtual bool IsBow()	// Or a Bow?
    { return FALSE; }
    virtual bool IsUnique()	// Anything else which look like this?
    { return FALSE; }
    virtual bool light()	// Is the object lit?
    { return FALSE; }
    virtual bool visible()	// Is the object visible?
    { return TRUE; }
    bool match(symbol_type type, char c); // Match symbol, letter?
    virtual long count()	// Number of items contained in this object
    { return 1L; }
    virtual comp compare(OBJECT* object); // Used to sort the objects
    virtual bool CanRemove(MONSTER*) // Is it cursed?
    { return FALSE; }
    virtual void TakeOff(MONSTER*) // Stop wearing it
    { }
    virtual STRING print(print_flag flag, long count, PLAYER*);
    virtual const char* name(); // Name of an object
    virtual const char* PluralName(); // Name of several objects
    virtual const char* ClassName(); // Name of the kind of objects
    virtual const char* PluralClassName(); // Name several object of the kind
    virtual long size()		// How large is it?
    { return 0L; }
    virtual long weight()	// How heavy is it?
    { return 0L; }
    virtual int strength()	// How easy is it to ruin?
    { return 1; }
    virtual long worth(MONSTER*)	// How much is it worth?
    { return 0L; }
    virtual int DamageBonus(MONSTER*, MONSTER*)
    { return 0; }
    virtual int HitBonus(MONSTER*, MONSTER*)
    { return 0; }
    virtual int ThrowBonus(MONSTER*, OBJECT*)
    { return 0; }
    virtual int DigBonus(MONSTER*)	
    { return strength(); }
    virtual damage_type DamageType() // Most objects just hit
    { return DAMAGE_HIT; }
    //
    // Use
    //
    virtual void apply(LEVEL*, MONSTER* monster); // General use by monster
    virtual QUESTION* DoApply(LEVEL* level, PLAYER* player, CHANNEL*);
    virtual bool MoveUp(LEVEL*, MONSTER*);
    virtual bool MoveDown(LEVEL*, MONSTER*);
    virtual bool AcceptEnter(MONSTER*);
    virtual bool AcceptLeave(MONSTER*);
    virtual QUESTION* GetQuestion(PLAYER*, LEVEL*, coor);
    virtual cont defend(MONSTER*, LEVEL*);
    virtual cont event(LEVEL*, coor, MONSTER* = NULL)
    { return CONTINUE; }
    virtual long NextEvent()
    { return 0L; }
    virtual void InsertAbility(MONSTER*) // Insert special ability in pack
    { }
    //
    // Violence
    //
    virtual OBJECT* polymorph(LEVEL*) // Change this object
    { return NULL; /* Default: polymorph failed. */ }
    virtual void MoveAway(LEVEL* level, coor pos, int);	// Teleport
    virtual void curse() { }	// Curse object when owner dies etc.
    virtual void bless() { }	// Bless object when ?
    virtual void SetVisible() { } // Make visible
    virtual void ClearVisible() { } // Make invisible
    virtual void RemoveCurse() { } // Remove curse && bless
    virtual void ChangeValue(int) { } // Change worth of object
    virtual void damage() { }	// Damage the object
    // Test if this object is likely to be hit
    virtual bool MightHit();
    // Try to catch something throw at you
    virtual bool catch(int, OBJECT*);
    // Test if it hit's
    virtual bool hit(int, OBJECT*, MONSTER*);
    // Hit, so be hurt!
    virtual STRING hurt(int power, OBJECT* object, MONSTER*, LEVEL*);
    // Absorb some of the damage
    virtual cont absorb(int*, damage_type*, OBJECT*, MONSTER*, MONSTER*)
    { return CONTINUE; } 
    //
    // Administration
    //
    virtual void SetLetter(char); // Set this objects inventory entry
    virtual void add(OBJECT*); // Add object to this
    virtual OBJECT* split(long); // Split this object in two
    virtual void save(BINARY* binary);
    OBJECT(BINARY* binary);
    OBJECT(PARSE*);
    OBJECT();
    virtual ~OBJECT() { }
};

extern OBJECT* LoadObject(BINARY* binary);
extern OBJECT* ParseObject(PARSE* parse, coor pos);
extern OBJECT* CreateObject(object_type type);

// @@ end

#endif OBJECT_HEADER

// @@ emacs
//
// Local Variables:
// mode: c++
// outline-regexp: "// @+ "
// eval: (outline-minor-mode 1)
// End:
