// @ myth - server - player.C
//
// Myth is a multiuser screen oriented adventure game.
// Copyright (C) 1989 Per Abrahamsen
// 
// This file is part of Myth.
// 
// Myth is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 1, or (at your option) any later
// version.
// 
// Myth is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Myth; see the file COPYING.  If not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// My email address is: amanda@iesd.auc.dk
//
// Definition of rutines used by the player object.

// @@ include

#include "server/player.h"
#include "common/ipc.h"
#include "common/error.h"
#include "common/parse.h"
#include "server/question.h"
#include "common/page.h"
#include "common/wait.h"
#include "server/sleep.h"
#include "server/watch.h"
#include "server/corpse.h"
#include "common/format.h"

// @@ const

// Level to put new players
static const level_type StartLevel = LEVEL_HOME;
// Position at level to put new players
static const char StartLocation = '<';
// 24 Hours energy
static const long NewEnergy = 24L * 60L * 60L / TickTime;

// @@ PLAYER

// @@@ STRING PLAYER::passwd()

STRING PLAYER::passwd()
{ return PassWord; }

// @@@ const char* PLAYER::LivingName()

const char* PLAYER::LivingName()
{
    if(MyName.length() > 0)
	return MyName();
    else
	return body->name();
}

// @@@ const char* PLAYER::name()

const char* PLAYER::name()
{
    if(MyName.length() > 0)
    {
	if(dead())
	{
	    STRING string = format("%s's dead body", MyName());
	    const char* text = string.string();
	    return text;
	}
	else
	    return MyName();
    }
    else
	return body->name();
}

// @@@ void PLAYER::ReName(STRING name)

void PLAYER::ReName(STRING name)
{ MyName = name; }

// @@@ void PLAYER::SetPasswd(STRING passwd)

void PLAYER::SetPasswd(STRING passwd)
{ PassWord = passwd; }

// @@@ void PLAYER::learn(OBJECT* object, int value = 1)

void PLAYER::learn(OBJECT* object, int value = 1)
{
    // Your pack might change
    if(knowing[object->type()] == 0)
    {
	SPECIFY_TYPE specify(object->type());
	if(pack.contain(&specify))
	    pack.touch();
    }
    knowing[object->type()] = min(100, knowing[object->type()] + value);
}

// @@@ void PLAYER::identify(OBJECT* object)

void PLAYER::identify(OBJECT* object)
{
    // Learn about the object, and its type
    known.add(object);
    learn(object);
    
    // And mark the pack as changed, if it belong there
    SPECIFY_OBJECT specify(object);
    if(pack.contain(&specify))
	pack.touch();
}

// @@@ void PLAYER::forget(int chance)

void PLAYER::forget(int chance)
{
    // Forget about a type of objects
    INVENTORY inventory;
    known.forget(chance, &inventory);
    pack.touch();
}

// @@@ bool PLAYER::ready()

bool PLAYER::ready()
{ return channel != NULL && channel->IsReady; }

// @@@ bool PLAYER::good()

bool PLAYER::good()
{
    if(channel == NULL)
	return FALSE;
    else if(channel->GoodInput() && channel->GoodOutput())
	return TRUE;
    else
	return FALSE;
}    

// @@@ void PLAYER::SendBunch(inventory_type type, PAGE* page)

void PLAYER::SendBunch(inventory_type type, PAGE* page)
{
    if(channel != NULL)
    {
	channel->PutChar(SERVER_INVENTORY);
	channel->PutChar(type);
	
	page->reset();
	
	while(page->remain() > 0)
	{
	    channel->PutString(page->string());
	    
	    page->next();
	}
    }
}

// @@@ void PLAYER::SpecialEffect(coor pos, char c)

void PLAYER::SpecialEffect(coor pos, char c)
{
    channel->PutChar(SERVER_EFFECT);
    channel->PutCoor(pos);
    channel->PutChar(c);
}

// @@@ void PLAYER::refresh()

void PLAYER::refresh()
{
    // Send new status of player
    //
    for(status_type type = STATUS_SERVER; type < STATUS_SERVER_END; type++)
    {
	if(OldStatus[type] != NewStatus[type])
	{
	    OldStatus[type] = NewStatus[type];
	    channel->PutChar(SERVER_STATUS);
	    channel->PutChar(type);
	    channel->PutLong(NewStatus[type]);
	}
    }
    //
    // Flush data if any
    //
    if(channel != NULL && channel->CheckTouched())
    {
	channel->PutChar(SERVER_READY);
	channel->flush();
	channel->ClearTouched();
    }
}

// @@@ void PLAYER::SendStatus(status_type type, status_value value)

void PLAYER::SendStatus(status_type type, status_value value)
{
    if(NewStatus[type] != value)
	NewStatus[type] = value;
}

// @@@ void PLAYER::SetWaiting()

void PLAYER::SetWaiting()
{
    if(!CheckWaiting())
    {
	SendStatus(STATUS_GAME, GAME_WAIT);
	refresh();
    }
}

// @@@ void PLAYER::ClearWaiting()

void PLAYER::ClearWaiting()
{
    if(CheckWaiting())
    {
	SendStatus(STATUS_GAME, GAME_PLAY);
	refresh();
    }
}

// @@@ void PLAYER::TickFight()

void PLAYER::TickFight()
{
    MONSTER::TickFight();
    ClearWaiting();
}

// @@@ void PLAYER::request(QUESTION* ask)

void PLAYER::request(QUESTION* ask)
{
    if(question)		// The BOOK says, if x is nil, delete x is nop
	delete question;

    question = ask;
}

// @@@ bool PLAYER::busy()

bool PLAYER::busy()
{ return question != NULL; }

// @@@ void PLAYER::parse()

void PLAYER::parse()
{ if(ready()) question->parse(this->channel); }

// @@@ question_value PLAYER::answer(LEVEL* level)

question_value PLAYER::answer(LEVEL* level)
{
    if(question == NULL)
	return Q_VALUE_ERROR;
    else
	return question->play(level, this);
}

// @@@ void PLAYER::retry()

void PLAYER::retry()
{
    if(question != NULL)
	request(question->next());
}

// @@@ QUESTION* PLAYER::command(LEVEL* level, const char* string)    

QUESTION* PLAYER::command(LEVEL* level, const char* string)
{
    
    if(strcmp(string, "version") == EQUAL)
	message(format("server: %s.", MythVersion));
    else if(strcmp(string, "forget") == EQUAL)
    {
	// You can not resist the temptation to mimic Hack...
	message("You begin to think about Maud...");
	view.forget(channel);
    }
    else if(strncmp(string, "message", 7) == EQUAL)
    {
	string += 7;

	while(*string == ' ')
	    string++;

	if(strcmp(string, "wall") == EQUAL)
	    IsVerboseWall = TRUE;
	else if(strcmp(string, "nowall") == EQUAL)
	    IsVerboseWall = FALSE;
	else if(strcmp(string, "get") == EQUAL)
	    IsVerboseGet = TRUE;
	else if(strcmp(string, "noget") == EQUAL)
	    IsVerboseGet = FALSE;
	else
	    message("Say what?");
    }
    else if(strcmp(string, "hiss") == EQUAL)
    {
	if(MayFight())
	    body->hiss(level, this);
	else
	    message("You take a deep breath.");
    }
    else if(strcmp(string, "breathe") == EQUAL)
    {
	if(MayFight())
	    return new QUESTION_BREATHE(channel);
	
	message("I'm not ready, yet...");
    }
    else if(strcmp(string, "eat") == EQUAL)
    {
	BUNCH* bunch = level->create(position());
	bunch->send(INVENTORY_OTHER, this);
	return new QUESTION_EAT(channel);
    }
    else if(strncmp(string, "say", 3) == EQUAL)
	level->talk(this, string + 3);
    else
	message(format("Unkown command: %s", string));

    return NULL;
}

// @@@ void PLAYER::message(STRING string)

void PLAYER::message(STRING string)
{
    if(channel != NULL)
    {
	// Something happened. Stop to examine what.
	SetInterrupt();
	
	channel->PutChar(SERVER_NEWS);
	channel->PutString(string());
    }
}

// @@@ void PLAYER::PageMessage(const char* header,

void PLAYER::PageMessage(const char* header, const char* prefix, PAGE* page)
{
    message(header);

    page->reset();
    while(page->remain())
    {
	message(format("%s%s", prefix, page->string()));
	page->next();
    }
}

// @@@ void PLAYER::error(const char* where, const char* what)

void PLAYER::error(const char* where, const char* what)
{
    message(format("CHAOS: %s: %s.", where, what));
    bug.error(where, what);
}

// @@@ void PLAYER::error(const char* where, STRING what)

void PLAYER::error(const char* where, STRING what)
{ error(where, what()); }

// @@@ void PLAYER::page(PAGE* page)

void PLAYER::page(PAGE* page)
{
    if(channel != NULL)
    {
	channel->PutChar(SERVER_PAGE);

	page->reset();
	while(page->remain() > 0)
	{
	    if(page->current().length() < 1)
		channel->PutString(" ");
	    else
		channel->PutString(page->string());

	    page->next();
	}
	channel->PutString("");
    }
}

// @@@ cont PLAYER::play(LEVEL*)

cont PLAYER::play(LEVEL*)
{
    return DONE;
}

// @@@ void PLAYER::enter(LEVEL* level)
//
// Let the player enter a new level in the dungeon

void PLAYER::enter(LEVEL* level)
{
    LastSeen = NoWhere;

    if(NewIdentity == NoIdentity) // Just changed level, or entered the game
    {
	if(channel != NULL)
	{
	    view.enter(depth(), channel);
	    SendStatus(STATUS_GAME, GAME_PLAY);
	}
	known.forget(1 + experience(), &pack);
	MONSTER::enter(level);	// Insert player among the monsters
	SendStatus(STATUS_DUNGEON, depth());
	
	update(level);

	level->AddPlayer(this);	// Insert player among the other players
    }
    else			// Trying to wake up a sleeping player
    {
	if(level->restore(NewIdentity, this) == SUCCES)
	    channel = NULL;	// Don't close this channel

	// Remove this player (a copy was created by LEVEL::restore)
	DungeonLevel = LEVEL_REMOVE;
    }
}

// @@@ void PLAYER::CreateMap(mapping_type type, int l)

void PLAYER::CreateMap(mapping_type type, int l)
{
    MapType = type;
    MapLength = l;
}

// @@@ void PLAYER::update(LEVEL* level)

void PLAYER::update(LEVEL* level)
{
    // Update counters
    //
    // This ought to happen only when he actually moves!
    //
    sleep(level, max(TickTime, 1));
    body->DigestFood(this);
    SpendTime(level);

    // Escape if no channel connected to player, or player moving or is dead
    if(channel == NULL || depth() != level->depth())
	return;

    // Create a new request, if none
    if(!busy())
	request(new QUESTION_COMMAND(channel));

    // If waiting for an answer (no automove), show him what happens meanwhile
    if(question->fresh() || position() != LastSeen)
    {
	// Update the view
	view.look(CheckAbility(ABILITY_BLIND) ? 1 : body->sight());
	view.detection(body->GetValue(ABILITY_DETECT));
	view.update(level, this, channel, SeeInvisible(), MapType, MapLength);
	CreateMap(MAPPING_SIGHT, 0);
	if(position() != LastSeen)
	{
	    // She always know how it looked where she left.
	    if(LastSeen != NoWhere)
		plot(level, LastSeen);
	    LastSeen = position();
	}
    }
    if(question->fresh())
    {
	// Send pack and body status to player iff needed
	pack.SendPack(this);
	body->status(this);
	
	// Send buffer to player
	refresh();
    }
}

// @@@ void PLAYER::plot(LEVEL*level, coor pos)

void PLAYER::plot(LEVEL*level, coor pos)
{ view.plot(level, pos, SeeInvisible(), channel); }

// @@@ void PLAYER::move(LEVEL* level, coor pos)

void PLAYER::move(LEVEL* level, coor pos)
{
    MONSTER::move(level, pos);

    if(channel != NULL)
	view.move(level, pos, channel);
}

// @@@ void PLAYER::teleport(LEVEL* level, int length,

void PLAYER::teleport(LEVEL* level, int length, holy_type holy = NORMAL)
{
    if(busy() || holy != BLESSED || GetChannel() == NULL)
	MONSTER::teleport(level, length, holy);
    else
	request(new QUESTION_TELEPORT(length, GetChannel()));
}

// @@@ void PLAYER::sleep(LEVEL*, long value)
//
// Let player sleep. This will cause him to leave the game.

void PLAYER::sleep(LEVEL*, long value)
{
    if(value == 0L)
	energy = -1L;
    else
	energy -= value;

    if(channel == NULL)
	return;
    else if(sleeping())
    {
	// Don't change location when restored
	NewLocation = ERROR;
	message("You fall asleep.");
    }
    else if(value > 60L * 60L)
	message("You feel sleepy.");
    else if(value < -60L)
	message("You fell fresh.");
    else switch(energy)
    {
    case 5L:
	message("You are falling to sleep.");
	break;
    case 30L:
	message("You are almost sleeping.");
	break;
    case 5L * 60L:
	message("You really need to sleep now.");
	break;
    case 30L * 60L:
	message("You are very tired.");
	break;
    case 60L * 60L:
	message("You are beginning to feel tired.");
    }
}

// @@@ void PLAYER::connect(CHANNEL* channel)

void PLAYER::connect(CHANNEL* channel)
{
    PLAYER::channel = channel;
    SendStatus(STATUS_EXPERIENCE, experience());
    body->status(this);
    pack.SendPack(this, FORCE);
    ClearWaiting();
}

// @@@ void PLAYER::DisConnect(WAIT* wait = NULL)

void PLAYER::DisConnect(WAIT* wait = NULL)
{
    if(channel != NULL)
    {
	if(body != NULL)
	    body->status(this);

	pack.SendPack(this);

	if(sleeping())
	    SendStatus(STATUS_GAME, GAME_SLEEP);
	else
	{
	    message(format("You %s with %ld point%s.",
			   rip(), MyScore, (MyScore == 1L) ? "" : "s"));
	    SendStatus(STATUS_GAME, GAME_OVER);
	}
	refresh();
	if(wait != NULL)
	    wait->remove(channel);
	else
	    bug.error("PLAYER::DisConnect", "wait == NULL");
	
	delete channel;
	channel = NULL;
    }
    if(question != NULL)	// Avoid core dumps
    {
	delete question;
	question = NULL;
    }
}    

// @@@ void PLAYER::restore(level_type depth, id_type id)

void PLAYER::restore(level_type depth, id_type id)
{
    DungeonLevel = depth;
    NewIdentity = id;
}

// @@@ void PLAYER::save(BINARY* binary)

void PLAYER::save(BINARY* binary)
{
    MONSTER::save(binary);
    view.save(binary);
    binary->PutString(MyName.length() > 0 ? MyName() : "");
    binary->PutChar(IsVerboseWall);
    binary->PutChar(IsVerboseGet);
    binary->PutLong(energy);
    binary->PutLong(watch());
    known.save(binary);
    for(int i = 0; i < OBJECT_END; i++)
	binary->PutChar(knowing[i]);
}

// @@@ void PLAYER::create()

void PLAYER::create()
{

    body = new BODY_HUMAN;
    body->SetExperience(experience());
    body->SetStrength(5);
    MyLetter = 'h';
    DungeonLevel = StartLevel;
    NewLocation = StartLocation;
    channel = NULL;
    question = NULL;
    for(status_type type = STATUS_SERVER; type < STATUS_SERVER_END; type++)
	OldStatus[type] = NewStatus[type] = FALSE;
    interrupt = FALSE;
    LastSeen = NoWhere;
    PassWord = STRING("");
    NewIdentity = NoIdentity;
    IsVerboseWall = FALSE;
    IsVerboseGet = TRUE;
    energy = NewEnergy;
    for(int i = 0; i < OBJECT_END; i++)
	knowing[i] = 0;
    CreateMap(MAPPING_SIGHT, 0);
}

// @@@ PLAYER::PLAYER(BINARY* binary)

PLAYER::PLAYER(BINARY* binary) : (binary)
{
    view.load(binary);
    channel = NULL;
    question = NULL;
    interrupt = FALSE;
    for(status_type type = STATUS_SERVER; type < STATUS_SERVER_END; type++)
	OldStatus[type] = NewStatus[type] = FALSE;
    LastSeen = NoWhere;
    ReName(binary->GetString());
    PassWord = STRING("");
    NewIdentity = NoIdentity;
    IsVerboseWall = binary->GetChar();
    IsVerboseGet = binary->GetChar();
    energy = binary->GetLong();
    energy = max(energy + (watch() - binary->GetLong()) / TickTime, NewEnergy);
    known.load(binary);
    for(int i = 0; i < OBJECT_END; i++)
	knowing[i] = binary->GetChar();
    CreateMap(MAPPING_SIGHT, 0);
}

// @@@ PLAYER::PLAYER(PARSE* parse, coor pos)

PLAYER::PLAYER(PARSE* parse, coor pos) : (parse, pos)
{
    create();

    for(;;)
    {
	const char* string = parse->GetWord();
	
	if(strcmp(string, "NAME") == EQUAL)
	{
	    parse->skip(INPUT_SPACE);
	    ReName(STRING(parse->GetWord()));
	}
	else if(strcmp(string, "ENERGY") == EQUAL)
	{
	    parse->skip(INPUT_SPACE);
	    energy = parse->GetLong();
	}
	else if(strcmp(string, "BREAK") == EQUAL)
	    break;
	else
	    parse->wrong("PLAYER::PLAYER",
			 format("Unknown field: %s", string));

	parse->skip(INPUT_COMMENT);

    }
    parse->skip(INPUT_COMMENT);
}

// @@@ PLAYER::PLAYER(GATE* gate, WAIT* wait, SLEEP* sleep)

PLAYER::PLAYER(GATE* gate, WAIT* wait, SLEEP* sleep)
{
    create();
    connect(new CHANNEL(gate));
    wait->add(channel);
    request(new QUESTION_START_UP(this, sleep));
    refresh();
}

// @@@ PLAYER::~PLAYER()

PLAYER::~PLAYER()
{ DisConnect(); }

// @@ emacs
//
// Local Variables:
// mode: c++
// outline-regexp: "// @+ "
// eval: (outline-minor-mode 1)
// End:

