// @ myth - server - player.h
//
// Myth is a multiuser screen oriented adventure game.
// Copyright (C) 1989 Per Abrahamsen
// 
// This file is part of Myth.
// 
// Myth is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 1, or (at your option) any later
// version.
// 
// Myth is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
// 
// You should have received a copy of the GNU General Public License
// along with Myth; see the file COPYING.  If not, write to the Free
// Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
//
// My email address is: amanda@iesd.auc.dk
//
// The player (as an object in the game) is defined in this file.

// @@ begin

#ifndef PLAYER_HEADER
#define PLAYER_HEADER

// @@ include

#include "server/monster.h"
#include "server/view.h"
#include "server/question.h"

struct GATE;
struct CHANNEL;
struct WAIT;


// @@ PLAYER

class PLAYER : public MONSTER
{
    // Identification
    //
private:
    STRING MyName;		// The players name
    STRING PassWord;		// Used when restored
    id_type NewIdentity;	// Restore as this player
public:
    object_type type()		// This is a player
    { return MONSTER_PLAYER; }
    bool IsPlayer()		// Identify object as player
    { return !dead(); }
    const char* LivingName();
    const char* name();		// What is his name?
    bool IsUnique()		// Only one player with this name.
    { return MyName.length() > 0; }
    STRING passwd();		// Return selected password
    void ReName(STRING name);	// Give player a (new) name and passwd
    void SetPasswd(STRING passwd); // Give player a new passwd
    int dexterity()		// How efficient to hit
    { return 15; }
    //
    // Knowledge
    //
private:
    ID_LIST known;		// List of all identified object instances
    int knowing[OBJECT_END];	// List of all identified object types
public:
    bool know(OBJECT* object)	// Is this object identified?
    { return knowledge(object) > 99 || known.member(object); }
    int knowledge(OBJECT* object) // Does he know this kind of object?
    { return knowing[object->type()]; }
    void learn(OBJECT* object, int value = 1); // Learn about an object type
    void identify(OBJECT* object); // Identify an instance af an object
    void forget(int chance); // Forget about a type of objects
    bool DragonKiller()	     // Has he killed a dragon yet?
    { return knowing[MONSTER_DRAGON] > 0; }
    //
    // The Channel
    //
private:
    CHANNEL* channel;		// The channel to the client process
public:
    CHANNEL* GetChannel()	// Avoid use of this
    { return channel; }
    bool ready();		// Data ready on channel
    bool good();		// Channel is working
    void SendBunch(inventory_type type, PAGE* page); // Update inventory
    void SpecialEffect(coor pos, char c); // Mark position on screen
    void refresh();		// Tell client we are ready
    //
    // Status
    //
private:
    status_value OldStatus[STATUS_END];	// Status as last shown to the client
    status_value NewStatus[STATUS_END];	// Current status in the game
public:
    void SendStatus(status_type type, status_value value); // Set client status
    void SetWaiting();
    bool CheckWaiting()
    { return NewStatus[STATUS_GAME] == GAME_WAIT; }
    void ClearWaiting();
    void TickFight();
    //
    // The Question
    //
private:
    QUESTION* question;		// The current question being answered
public:
    void request(QUESTION* ask); // Ask a question
    bool busy();	    	// TRUE, if player is answering a question
    void parse();		// Parse input from the channel
    question_value answer(LEVEL* level); // Use the request
    void retry();		// Create a follow-up to the answer
    QUESTION* command(LEVEL* level, const char* string); // Extended commands
    //
    // Messages
    //
private:
    bool IsVerboseWall;		// Control some messages.
    bool IsVerboseGet;		// Control some other messages.
public:
    inline bool VerboseWall()
    { return IsVerboseWall; }
    inline bool VerboseGet()
    { return IsVerboseGet; }
    inline void message(const char* string)
    { message(STRING(string)); }
    void message(STRING string); //  Send a message to the news handler
    void PageMessage(const char* header, const char* prefix, PAGE* page);
    void error(const char* where, const char* what); // Tell error to player
    void error(const char* where, STRING what);
    void page(PAGE* page);	// Send page to player
    //
    // Interrupt
    //
private:
    bool interrupt;		// TRUE, iff the player shouldn't automove
public:
    inline void SetInterrupt()
    { interrupt = TRUE; }
    inline bool CheckInterrupt()
    { return interrupt; }
    inline void ClearInterrupt()
    { interrupt = FALSE; }
    //
    // View
    //
private:
    VIEW view;			// Handle the view of the player
    mapping_type MapType;	// A magic mapping?
    int MapLength;
    coor LastSeen;		// Last position viewed by player
    bool IsTouched;		// Is any output written to the player?
public:
    void CreateMap(mapping_type type, int l);
    bool CanSee(coor where)
    { return view.CanSee(where); }
    void update(LEVEL* level);	// Update the players screen & pack
    void plot(LEVEL* level, coor pos); // Examine a position
    //
    // Action
    //
private:
    long energy;		// How long time, until he fall asleep
public:
    cont play(LEVEL* level);	// Always finished moving.
    void enter(LEVEL* level);	// Enter new level
    void move(LEVEL* level, coor pos); // Move to a new position
    void teleport(LEVEL* level, int length, holy_type holy = NORMAL);
    void sleep(LEVEL*, long value = 0L); // You get tired...
    inline bool sleeping()	// Check if player is sleeping
    { return energy < 1L; }
    //
    // Administration
    //
private:
    void create();		// Initialise object.
public:
    void connect(CHANNEL* channel); // Connect player to channel
    void DisConnect(WAIT* wait = NULL); // Disconnect player from channel
    void restore(level_type depth, id_type id);	// restore old game
    void save(BINARY* binary);
    PLAYER(BINARY* binary);
    PLAYER(GATE* gate, WAIT* wait, SLEEP* sleep);
    PLAYER(PARSE* parse, coor pos);
    ~PLAYER();
};

// @@ end

#endif PLAYER_HEADER

// @@ emacs
//
// Local Variables:
// mode: c++
// outline-regexp: "// @+ "
// eval: (outline-minor-mode 1)
// End:
